"""Path module.

For handling local files and directory operations.
"""

from pathlib import Path

import click

from imr.imr import IMRLocal, IMRRemote, load_config


class Context:
    """The context for all CLI commands."""

    home = Path.home()
    imr_dir = home / ".imr"
    imr_local: IMRLocal
    imr_remote: IMRRemote


@click.group()
@click.pass_context
def cli(ctx: click.Context) -> None:
    """Get the cli command options."""
    ctx.obj = Context()


@cli.group()
@click.option(
    "-c", "--config", type=str, help="name of the configuration."
)
@click.pass_obj
def local(obj: Context, config: str) -> None:
    """Get local command options."""
    if config is not None:
        config_params = load_config(config)
        obj.imr_local = IMRLocal(config_params["path"])
    else:
        obj.imr_local = IMRLocal(str(obj.imr_dir))


@local.command("list")
@click.pass_obj
def list_local(obj: Context) -> None:
    """List local packages."""
    for package in obj.imr_local.list():
        click.echo(package)


@local.command("rm")
@click.argument("package")
@click.option(
    "-v", "--version", type=str, default="latest", help="version of the model.", show_default=True
)
@click.pass_obj
def rm_local(obj: Context, package: str, version: str) -> None:
    """Remove local packages."""
    obj.imr_local.rm(package, version)


@local.command("path")
@click.argument("package")
@click.option(
    "-v", "--version", type=str, default="latest", help="version of the model.", show_default=True
)
@click.pass_obj
def path_local(obj: Context, package: str, version: str) -> None:
    """Get remote model path."""
    click.echo(obj.imr_local.path(package, version))


@cli.group()
@click.option(
    "-h", "--host", type=str, help="Server host address ie. http://mymodelhost.me."
)
@click.option(
    "-u", "--user", type=str, help="Server host user."
)
@click.option(
    "-p", "--password", type=str, help="Server host password."
)
@click.option(
    "-c", "--config", type=str, help="Remote configuration."
)
@click.pass_obj
def remote(obj: Context, host: str, user: str, password: str, config: str) -> None:
    """Get remote command cli options."""
    if host is not None:
        obj.imr_remote = IMRRemote(host, user, password)
    else:
        config_params = load_config(config)
        obj.imr_remote = IMRRemote(config_params["host"], config_params["user"],
                                   config_params["password"])


@remote.command("list")
@click.pass_obj
def list_remote(obj: Context) -> None:
    """List remote packages."""
    packages = obj.imr_remote.list()
    for p in packages:
        click.echo(p)


@remote.command("rm")
@click.argument("package")
@click.option(
    "-v", "--version", type=str, default="latest", help="version of the model.", show_default=True
)
@click.pass_obj
def rm_remote(obj: Context, package: str, version: str) -> None:
    """Remove remote package."""
    obj.imr_remote.rm(package, version)


@remote.command("push")
@click.argument("model_dir")
@click.argument("package")
@click.option(
    "-v", "--version", type=str, default="latest", help="version of the model.", show_default=True
)
@click.pass_obj
def push_remote(obj: Context, model_dir: str, package: str, version: str) -> None:
    """Push model to remote repository."""
    obj.imr_remote.push(model_dir, package, version)


@remote.command()
@click.argument("package")
@click.option(
    "-d",
    "--dir",
    type=str,
    help="directory to pull the model in.",
    show_default=True,
)
@click.option(
    "-v", "--version", type=str, default="latest", help="version of the model.", show_default=True
)
@click.pass_obj
def pull(obj: Context, package: str, model_dir: str, version: str) -> None:
    """Pull model from remote repository."""
    obj.imr_remote.pull(model_dir, package, version)


@remote.command("path")
@click.argument("package")
@click.option(
    "-v", "--version", type=str, default="latest", help="version of the model.", show_default=True
)
@click.pass_obj
def path_remote(obj: Context, package: str, version: str) -> None:
    """Get remote model path."""
    click.echo(obj.imr_remote.path(package, version))



"""Intelligent Model Registry (imr)"""
import configparser
import shutil
from collections.abc import MutableMapping
from pathlib import Path

from artifactory import ArtifactoryPath
from requests.auth import HTTPBasicAuth

default_local_path = Path.home() / ".imr"
default_file_name = "imr.conf"


def load_config(config_name: str) -> MutableMapping:
    """Load configuration"""
    return load_config_from_file(default_local_path / default_file_name, config_name)


def load_config_from_file(file_path: Path, config: str) -> MutableMapping:
    """Load configuration from file."""
    config_params = configparser.ConfigParser()
    if file_path.exists():
        config_params.read(file_path)
    else:
        raise FileProcessingError(str(file_path), "Configuration file does not exist.")
    if config not in config_params:
        raise FileProcessingError(str(file_path), "Configuration does not exist: " + config)
    return config_params[config]


class FileProcessingError(Exception):
    """File does not exists Error."""

    def __init__(self, filename: str, message: str):
        self.filename = filename
        self.message = message
        super().__init__(f"Error processing file '{filename}': {message}")


class IMRRemote:
    """Remote repository class."""

    def __init__(self, repo: str, user: str, password: str):
        self.repo = repo
        self.user = user
        self.password = password
        self.home = str(Path.home())

    def list(self) -> list:
        """List models from the remote repository."""
        path = ArtifactoryPath(self.repo, auth=(self.user, self.password), auth_type=HTTPBasicAuth)
        return [str(package).replace(self.repo + "/", "") for package in path.glob("*/*")]

    def push(self, directory: str, package: str, version: str = "latest") -> None:
        """Push a local model to the remote repository."""
        path = ArtifactoryPath(
            self.repo + "/" + package + "/" + version,
            auth=(self.user, self.password),
            auth_type=HTTPBasicAuth,
        )
        path.mkdir()
        shutil.make_archive("model", "zip", directory)
        path.deploy_file("model.zip")

    def pull(self, directory: str, package: str, version: str = "latest") -> None:
        """Pull a remote model to the local repository."""
        path = ArtifactoryPath(
            self.repo + "/" + package + "/" + version + "/model.zip",
            auth=(self.user, self.password),
            auth_type=HTTPBasicAuth,
        )
        file_path: Path = Path(directory) / package / version
        if not Path(file_path).exists():
            Path.mkdir(file_path, parents=True)
        with path.open() as fd, Path.open(file_path / "model.zip", "wb") as out:
            out.write(fd.read())
        shutil.unpack_archive(file_path / "model.zip", file_path / "model")

    def rm(self, package: str, version: str = "latest") -> None:
        """Remove a model from the remote repository."""
        artefact = self.repo + "/" + package
        if version is not None:
            artefact = self.repo + "/" + package + "/" + version
        path = ArtifactoryPath(artefact, auth=(self.user, self.password), auth_type=HTTPBasicAuth)
        if path.exists():
            path.unlink()

    def path(self, package: str, version: str = "latest") -> ArtifactoryPath:
        """Return the path to the model."""
        artefact = self.repo + "/" + package
        if version is not None:
            artefact += "/" + version
        return ArtifactoryPath(artefact, auth=(self.user, self.password), auth_type=HTTPBasicAuth)


class IMRLocal:
    """Local repository class."""

    repo: Path
    home: Path

    def __init__(self, repo: str | None = None):
        self.home = Path.home()
        if repo is None:
            self.repo = self.home / ".imr"
        else:
            self.repo = Path.expanduser(Path(repo))
        if not Path(self.repo).exists():
            Path(self.repo).mkdir(parents=True)

    def list(self) -> list[str]:
        """List local models.

        returns all the paths in the repository after main directory
        BASE/modela/version1
        BASE/modela/version2
        BASE/modelb/version3

        results in ["modela/version1", "modela/version2", "modelb/version3"]
        """
        return ["/".join(version.parts[-2:]) for version in self.repo.rglob("*/*/")]

    def push(self, directory: str, package: str, version: str = "latest") -> None:
        """Push a model in a directory to the local repository."""
        shutil.copytree(src=directory, dst=self.repo / package / version)

    def rm(self, package: str, version: str = "latest") -> None:
        """Remove a model from the local repository."""
        if version is None:
            shutil.rmtree(self.repo / package)
        else:
            shutil.rmtree(self.repo / package / version)

    def path(self, package: str, version: str = "latest") -> Path:
        """Return the path to the model."""
        path: Path = self.repo / package
        if version is not None:
            path = self.repo / package / version
        return path


def imr_from_config(config_name: str) -> IMRRemote | IMRLocal:
    """Create a repository based on the given configuration name."""
    return imr_from_file(default_local_path / default_file_name, config_name)


def imr_from_file(file_path: Path, config_name: str) -> IMRRemote | IMRLocal:
    """Create a repository based on the given configuration name."""
    repo_config = load_config_from_file(file_path, config_name)
    if "type" not in repo_config:
        raise FileProcessingError("imr.conf", "Repository type not specified.")
    if repo_config["type"] == "local":
        repo_path = repo_config["path"]
        return IMRLocal(repo_path)
    if repo_config["type"] == "remote":
        repo_path = repo_config["path"]
        user = repo_config["user"]
        password = repo_config["password"]
        return IMRRemote(repo_path, user, password)
    raise FileProcessingError("imr.conf", "Repository type not known: " + repo_config["type"])
